# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.date import get_date_resolution, parse_date
from aniso8601.duration import get_duration_resolution, parse_duration
from aniso8601.interval import (
    get_interval_resolution,
    get_repeating_interval_resolution,
    parse_interval,
    parse_repeating_interval,
)

# Import the main parsing functions so they are readily available
from aniso8601.time import (
    get_datetime_resolution,
    get_time_resolution,
    parse_datetime,
    parse_time,
)

__version__ = "10.0.0"
